# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""
DMARC urls
https://dmarc.org/resources/specification/
"""
from django.conf.urls import url

from dmarc import views

app_name = 'dmarc'
urlpatterns = [
    url("^report/$", views.dmarc_report, name='dmarc_report'),
    url("^report/csv/$", views.dmarc_csv, name='dmarc_csv'),
    url("^report/json/$", views.dmarc_json, name='dmarc_json'),
]



# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""
DMARC models for managing Aggregate Reports
https://dmarc.org/resources/specification/
"""
from django.db import models


class Reporter(models.Model):
    """DMARC reporter"""

    org_name = models.CharField('Organisation', unique=True, max_length=100)
    email = models.EmailField()

    def __unicode__(self):
        return self.org_name


class Report(models.Model):
    """DMARC report metadata"""

    report_id = models.CharField(max_length=100)
    reporter = models.ForeignKey(Reporter, on_delete=models.CASCADE)
    date_begin = models.DateTimeField(db_index=True)
    date_end = models.DateTimeField()
    policy_domain = models.CharField(max_length=100)
    policy_adkim = models.CharField('DKIM alignment mode', max_length=1)
    policy_aspf = models.CharField('SPF alignment mode', max_length=1)
    policy_p = models.CharField('Requested handling policy', max_length=10)
    policy_sp = models.CharField('Requested handling policy for subdomains', max_length=10)
    policy_pct = models.SmallIntegerField('Sampling rate')
    report_xml = models.TextField(blank=True)

    def __unicode__(self):
        return self.report_id

    class Meta(object):
        """Model constraints"""
        # pylint: disable=too-few-public-methods

        unique_together = (("reporter", "report_id", "date_begin"),)


class Record(models.Model):
    """DMARC report record"""

    report = models.ForeignKey(Report, related_name='records', on_delete=models.CASCADE)
    source_ip = models.CharField(max_length=39)
    recordcount = models.IntegerField()
    policyevaluated_disposition = models.CharField(max_length=10)
    policyevaluated_dkim = models.CharField(max_length=4)
    policyevaluated_spf = models.CharField(max_length=4)
    policyevaluated_reasontype = models.CharField(blank=True, max_length=75)
    policyevaluated_reasoncomment = models.CharField(blank=True, max_length=100)
    identifier_headerfrom = models.CharField(max_length=100)

    def __unicode__(self):
        return self.source_ip


class Result(models.Model):
    """DMARC report record result"""

    record = models.ForeignKey(Record, related_name='results', on_delete=models.CASCADE)
    record_type = models.CharField(max_length=4)
    domain = models.CharField(max_length=100)
    result = models.CharField(max_length=9)

    def __unicode__(self):
        return "%s:%s-%s" % (str(self.pk), self.record_type, self.domain,)


class FBReporter(models.Model):
    """DMARC feedback reporter"""

    org_name = models.CharField('Organisation', unique=True, max_length=100)
    email = models.EmailField()

    def __unicode__(self):
        return self.email

    def save(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        if not self.org_name:
            self.org_name = self.email
        super(FBReporter, self).save(*args, **kwargs)


class FBReport(models.Model):
    """DMARC feedback report"""

    reporter = models.ForeignKey(FBReporter, on_delete=models.CASCADE)
    date = models.DateTimeField(db_index=True)
    source_ip = models.CharField(max_length=39)
    domain = models.CharField(max_length=100)
    email_from = models.CharField(max_length=100, blank=True)
    email_subject = models.CharField(max_length=100, blank=True)
    spf_alignment = models.CharField(max_length=10, blank=True)
    dkim_alignment = models.CharField(max_length=10, blank=True)
    dmarc_result = models.CharField(max_length=10, blank=True)
    description = models.TextField('human readable feedback', blank=True)
    email_source = models.TextField('source email including rfc822 headers', blank=True)
    feedback_report = models.TextField(blank=True)
    feedback_source = models.TextField()

    def __unicode__(self):
        msg = '{} {} {} {} {}'.format(
            self.date,
            self.domain,
            self.source_ip,
            self.email_from,
            self.email_subject
        )
        return msg



# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""Import DMARC Aggregate Reports"""
from __future__ import unicode_literals

import gzip
import logging
import os
import tempfile
import xml.etree.ElementTree as ET
import zipfile
from argparse import FileType
from datetime import datetime
from email import message_from_string

import pytz
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.db import Error
from django.db.utils import IntegrityError
from six import BytesIO

from dmarc.models import Record, Report, Reporter, Result


class Command(BaseCommand):
    """
    Command class for importing DMARC Aggregate Reports
    Most errors are not raised to prevent email bounces
    """
    help = 'Imports a DMARC Aggregate Report from either email or xml'

    def add_arguments(self, parser):
        parser.add_argument(
            '-e',
            '--email',
            type=FileType('r'),
            default=False,
            help='Import from email file, or - for stdin'
        )
        parser.add_argument(
            '-x',
            '--xml',
            type=FileType('r'),
            default=False,
            help='Import from xml file, or - for stdin'
        )

    def handle(self, *args, **options):
        """
        Handle method to import a DMARC Aggregate Reports
        Either pass in
        - the email message and the DMARC XML data will be extracted;
        - or the xml file.
        """
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        logger = logging.getLogger(__name__)
        logger.info("Importing DMARC Aggregate Reports")

        dmarc_xml = ''

        if options['email']:
            email = options['email'].read()
            msg = 'Importing from email: {}'.format(email)
            dmarc_xml = self.get_xml_from_email(email)
        elif options['xml']:
            dmarc_xml = options['xml'].read()
            msg = 'Importing from xml: {}'.format(dmarc_xml)
            logger.debug(msg)
        else:
            msg = "Check usage, please supply a single DMARC report file or email"
            logger.error(msg)
            raise CommandError(msg)

        tz_utc = pytz.timezone('UTC')
        try:
            root = ET.fromstring(dmarc_xml)
        except:
            msg = "Processing xml failed: {}".format(dmarc_xml)
            logger.error(msg)
            return

        # Report metadata
        report_metadata = root.findall('report_metadata')
        org_name = None
        email = None
        report_id = None
        report_begin = None
        report_end = None
        for node in report_metadata[0]:
            if node.tag == 'org_name':
                org_name = node.text
            if node.tag == 'email':
                email = node.text
            if node.tag == 'report_id':
                report_id = node.text
            if node.tag == 'date_range':
                report_begin = node.find('begin').text
                report_end = node.find('end').text

        if org_name is None:
            msg = "This DMARC report does not have an org_name"
            logger.error(msg)
        if report_id is None:
            msg = "This DMARC report for {} does not have a report_id".format(org_name)
            logger.error(msg)
        try:
            reporter = Reporter.objects.get(org_name=org_name)
        except ObjectDoesNotExist:
            try:
                reporter = Reporter.objects.create(org_name=org_name, email=email)
            except Error as err:
                msg = "Unable to create DMARC report for {}: {}".format(org_name, err)
                logger.error(msg)

        # Reporting policy
        policy_published = root.findall('policy_published')
        # Set defaults
        policy_domain = None
        policy_adkim = 'r'
        policy_aspf = 'r'
        policy_p = 'none'
        policy_sp = 'none'
        policy_pct = 0
        for node in policy_published[0]:
            if node.tag == 'domain':
                policy_domain = node.text
            if node.tag == 'adkim':
                policy_adkim = node.text
            if node.tag == 'aspf':
                policy_aspf = node.text
            if node.tag == 'p':
                policy_p = node.text
            if node.tag == 'sp':
                policy_sp = node.text
            if node.tag == 'pct':
                policy_pct = int(node.text)

        # Create the report
        report = Report()
        report.report_id = report_id
        report.reporter = reporter
        report_date_begin = datetime.fromtimestamp(float(report_begin)).replace(tzinfo=tz_utc)
        try:
            report_date_begin = datetime.fromtimestamp(float(report_begin)).replace(tzinfo=tz_utc)
            report_date_end = datetime.fromtimestamp(float(report_end)).replace(tzinfo=tz_utc)
        except:
            msg = "Unable to understand DMARC reporting dates"
            logger.error(msg)
        report.date_begin = report_date_begin
        report.date_end = report_date_end
        report.policy_domain = policy_domain
        report.policy_adkim = policy_adkim
        report.policy_aspf = policy_aspf
        report.policy_p = policy_p
        report.policy_sp = policy_sp
        report.policy_pct = policy_pct
        report.report_xml = dmarc_xml
        try:
            report.save()
        except IntegrityError as err:
            msg = "DMARC duplicate report record: {}".format(err)
            logger.error(msg)
            return
        except Error as err:
            msg = "Unable to save the DMARC report header {}: {}".format(report_id, err)
            logger.error(msg)

        # Record
        for node in root.findall('record'):
            source_ip = None
            recordcount = 0
            policyevaluated_disposition = None
            policyevaluated_dkim = None
            policyevaluated_spf = None
            policyevaluated_reasontype = ''
            policyevaluated_reasoncomment = ''
            identifier_headerfrom = None
            row = node.find('row')
            source_ip = row.find('source_ip').text
            if row.find('count') is not None:
                recordcount = int(row.find('count').text)
            else:
                recordcount = 0
            policyevaluated = row.find('policy_evaluated')
            policyevaluated_disposition = policyevaluated.find('disposition').text
            policyevaluated_dkim = policyevaluated.find('dkim').text
            policyevaluated_spf = policyevaluated.find('spf').text
            if policyevaluated.find('reason') is not None:
                reason = policyevaluated.find('reason')
                if reason.find('type') is not None:
                    policyevaluated_reasontype = reason.find('type').text
                if reason.find('comment') is not None:
                    if reason.find('comment').text is not None:
                        policyevaluated_reasoncomment = reason.find('comment').text

            identifiers = node.find('identifiers')
            identifier_headerfrom = identifiers.find('header_from').text

            if not source_ip:
                msg = "DMARC report record useless without a source ip"
                logger.error(msg)

            # Create the record
            record = Record()
            record.report = report
            record.source_ip = source_ip
            record.recordcount = recordcount
            record.policyevaluated_disposition = policyevaluated_disposition
            record.policyevaluated_dkim = policyevaluated_dkim
            record.policyevaluated_spf = policyevaluated_spf
            record.policyevaluated_reasontype = policyevaluated_reasontype
            record.policyevaluated_reasoncomment = policyevaluated_reasoncomment
            record.identifier_headerfrom = identifier_headerfrom
            try:
                record.save()
            except IntegrityError as err:
                msg = "DMARC duplicate record: {}".format(err)
                logger.error(msg)
            except Error as err:
                msg = "Unable to save the DMARC report record: {}".format(err)
                logger.error(msg)

            auth_results = node.find('auth_results')
            for resulttype in auth_results:
                result_domain = resulttype.find('domain').text
                if result_domain is None:
                    # Allow for blank domains
                    result_domain = ''
                result_result = resulttype.find('result').text

                # Create the record
                result = Result()
                result.record = record
                result.record_type = resulttype.tag
                result.domain = result_domain
                result.result = result_result
                try:
                    result.save()
                except Error as err:
                    msg = "Unable to save the DMARC report result {} for {}: {}".format(
                        resulttype.tag,
                        result_domain,
                        err.message
                    )
                    logger.error(msg)

    @staticmethod
    def get_xml_from_email(email):
        """Get xml from an email"""
        # pylint: disable=too-many-statements
        dmarc_xml = ''
        logger = logging.getLogger(__name__)

        msg = 'Processing email'
        logger.debug(msg)
        try:
            dmarcemail = message_from_string(email)
        except:
            msg = 'Unable to use email'
            logger.debug(msg)
            return ''

        for mimepart in dmarcemail.walk():
            msg = 'Processing content type: {}'.format(mimepart.get_content_type())
            logger.debug(msg)
            if mimepart.get_content_type() in (
                    'application/x-zip-compressed',
                    'application/x-zip',
                    'application/zip',
                    'application/gzip',
                    'application/octet-stream'
            ):
                dmarc_zip = BytesIO()
                dmarc_zip.write(mimepart.get_payload(decode=True))
                dmarc_zip.seek(0)
                if zipfile.is_zipfile(dmarc_zip):
                    msg = "DMARC is zipfile"
                    logger.debug(msg)
                    try:
                        archive = zipfile.ZipFile(dmarc_zip, 'r')
                        files = archive.infolist()
                        # The DMARC report should only contain a single xml file
                        for file_ in files:
                            dmarc_xml = archive.read(file_)
                        archive.close()
                    except zipfile.BadZipfile:
                        msg = 'Unable to unzip mimepart'
                        logger.error(msg)
                        temp = tempfile.mkstemp(prefix='dmarc-', suffix='.zip')
                        dmarc_zip.seek(0)
                        tmpf = os.fdopen(temp[0], 'w')
                        tmpf.write(dmarc_zip.getvalue())
                        tmpf.close()
                        msg = 'Saved in: {}'.format(temp[1])
                        logger.debug(msg)
                        raise CommandError(msg)
                else:
                    msg = "DMARC trying gzip"
                    logger.debug(msg)
                    # Reset zip file
                    dmarc_zip.seek(0)
                    try:
                        archive = gzip.GzipFile(None, 'rb', 0, dmarc_zip)
                        dmarc_xml = archive.read()
                        archive = None
                        msg = "DMARC successfully extracted xml from gzip"
                        logger.debug(msg)
                    except:
                        msg = 'Unable to gunzip mimepart'
                        logger.error(msg)
                        temp = tempfile.mkstemp(prefix='dmarc-', suffix='.gz')
                        dmarc_zip.seek(0)
                        tmpf = os.fdopen(temp[0], 'w')
                        tmpf.write(dmarc_zip.getvalue())
                        tmpf.close()
                        msg = 'Saved in: {}'.format(temp[1])
                        logger.debug(msg)
                        raise CommandError(msg)
            else:
                try:
                    myname = mimepart.get_filename()
                except:
                    myname = 'Not provided'
                msg = "DMARC Report is not in mimepart: {}".format(myname)
                logger.debug(msg)
        return dmarc_xml



# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""
DMARC views
https://dmarc.org/resources/specification/
"""
import csv
import datetime

from django.contrib.admin.views.decorators import staff_member_required
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.db import connection
from django.http import JsonResponse, StreamingHttpResponse
from django.shortcuts import render

from dmarc.models import Report


class Echo(object):
    """An object that implements just the write method of the file-like
    interface for csv.writer.
    """
    # pylint: disable=too-few-public-methods

    @staticmethod
    def write(_, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


def _sql_cursor(request_args):
    """Returns a cursor according to users request"""
    # pylint: disable=too-many-statements
    sql_where = []
    sql_orderby = []
    sql_params = []
    if 'dmarc_date_from' in request_args:
        val = request_args['dmarc_date_from']
        try:
            val = datetime.datetime.strptime(val, '%Y-%m-%d')
        except:
            val = datetime.date.today()
        sql_where.append('dmarc_report.date_begin >= %s')
        sql_params.append(val)
    if 'dmarc_date_to' in request_args:
        val = request_args['dmarc_date_to']
        try:
            val = datetime.datetime.strptime(val, '%Y-%m-%d')
        except:
            val = datetime.date.today()
        val += datetime.timedelta(days=1)
        sql_where.append('dmarc_report.date_end < %s')
        sql_params.append(val)
    if 'dmarc_disposition' in request_args and request_args['dmarc_disposition']:
        val = request_args['dmarc_disposition']
        sql_where.append('dmarc_record.policyevaluated_disposition = %s')
        sql_params.append(val)
    if 'dmarc_onlyerror' in request_args:
        clause = '('
        clause += "dmarc_record.policyevaluated_dkim = 'fail'"
        clause += " OR "
        clause += "dmarc_record.policyevaluated_spf = 'fail'"
        clause += ')'
        sql_where.append(clause)
    if 'dmarc_filter' in request_args and request_args['dmarc_filter']:
        val = request_args['dmarc_filter'] + '%'
        clause = '('
        clause += "lower(dmarc_reporter.org_name) LIKE lower(%s)"
        clause += " OR "
        clause += "dmarc_record.source_ip LIKE %s"
        clause += ')'
        sql_where.append(clause)
        sql_params.append(val)
        sql_params.append(val)

    sql = """
SELECT
  dmarc_reporter.org_name,
  dmarc_reporter.email,
  dmarc_report.date_begin,
  dmarc_report.date_end,
  dmarc_report.policy_domain,
  dmarc_report.policy_adkim,
  dmarc_report.policy_aspf,
  dmarc_report.policy_p,
  dmarc_report.policy_sp,
  dmarc_report.policy_pct,
  dmarc_report.report_id,
  dmarc_record.source_ip,
  dmarc_record.recordcount,
  dmarc_record.policyevaluated_disposition,
  dmarc_record.policyevaluated_dkim,
  dmarc_record.policyevaluated_spf,
  dmarc_record.policyevaluated_reasontype,
  dmarc_record.policyevaluated_reasoncomment,
  dmarc_record.identifier_headerfrom,
  spf_dmarc_result.record_type AS spf_record_type,
  spf_dmarc_result.domain AS spf_domain,
  spf_dmarc_result.result AS spf_result,
  dkim_dmarc_result.record_type AS dkim_record_type,
  dkim_dmarc_result.domain AS dkim_domain,
  dkim_dmarc_result.result AS dkim_result
FROM dmarc_reporter
INNER JOIN dmarc_report
ON dmarc_report.reporter_id = dmarc_reporter.id
INNER JOIN  dmarc_record
ON dmarc_record.report_id = dmarc_report.id
LEFT OUTER JOIN dmarc_result AS spf_dmarc_result
ON spf_dmarc_result.record_id = dmarc_record.id
AND spf_dmarc_result.record_type = 'spf'
LEFT OUTER JOIN dmarc_result AS dkim_dmarc_result
ON dkim_dmarc_result.record_id = dmarc_record.id
AND dkim_dmarc_result.record_type = 'dkim'
    """

    if sql_where:
        sql = sql + " WHERE " + "\nAND ".join(sql_where)

    sql_orderby.append('LOWER(dmarc_reporter.org_name)')
    sql_orderby.append('dmarc_report.date_begin')
    sql_orderby.append('dmarc_record.source_ip')
    sql = sql + "\nORDER BY " + ", ".join(sql_orderby)

    cursor = connection.cursor()
    cursor.execute(sql, sql_params)

    return cursor


@staff_member_required
def dmarc_index(request):
    """Index view"""
    context = {
        "reports": 'TODO',
    }
    return render(request, 'dmarc/report.html', context)


@staff_member_required
def dmarc_report(request):
    """Paginated DMARC report list"""
    report_list = Report.objects.select_related(
        'reporter',
    ).prefetch_related(
        'records__results'
    ).order_by('-date_begin', 'reporter__org_name').all()
    paginator = Paginator(report_list, 2)

    page = request.GET.get('page')
    try:
        reports = paginator.page(page)
    except PageNotAnInteger:
        # If page is not an integer, deliver first page.
        reports = paginator.page(1)
    except EmptyPage:
        # If page is out of range (e.g. 9999), deliver last page of results.
        reports = paginator.page(paginator.num_pages)

    context = {
        "reports": reports,
    }
    return render(request, 'dmarc/report.html', context)


@staff_member_required
def dmarc_csv(request):
    """Export DMARC data as CSV"""
    # Inspired by https://code.djangoproject.com/ticket/21179
    def stream():
        """Generator function to yield cursor rows."""
        pseudo_buffer = Echo()
        writer = csv.writer(pseudo_buffer)
        columns = True
        for row in cursor.fetchall():
            data = ''
            if columns:
                # Write the columns if this is the first row
                columns = [col[0] for col in cursor.description]
                data = writer.writerow(columns)
                columns = False
            data = data + writer.writerow(row)
            yield data

    now = datetime.datetime.now()
    disposition = 'attachment; filename="dmarc-{}.csv"'.format(now.strftime('%Y-%m-%d-%H%M%S'))

    cursor = _sql_cursor(request.GET)

    response = StreamingHttpResponse(stream(), content_type="text/csv")
    response['Content-Disposition'] = disposition

    return response


@staff_member_required
def dmarc_json(request):
    """Export DMARC data as JSON"""
    cursor = _sql_cursor(request.GET)

    data = JsonResponse(cursor.fetchall(), safe=False)
    return data



# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""Django application configuration"""
from __future__ import unicode_literals

from django.apps import AppConfig


class DmarcConfig(AppConfig):
    """DMARC application configuration"""

    name = 'dmarc'
    verbose_name = "DMARC feedback report manager"



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-06-18 22:57
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FBReport',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(db_index=True)),
                ('source_ip', models.CharField(max_length=39)),
                ('domain', models.CharField(max_length=100)),
                ('email_from', models.CharField(blank=True, max_length=100)),
                ('email_subject', models.CharField(blank=True, max_length=100)),
                ('spf_alignment', models.CharField(blank=True, max_length=10)),
                ('dkim_alignment', models.CharField(blank=True, max_length=10)),
                ('dmarc_result', models.CharField(blank=True, max_length=10)),
                ('description', models.TextField(blank=True, verbose_name='human readable feedback')),
                ('email_source', models.TextField(blank=True, verbose_name='source email including rfc822 headers')),
                ('feedback_report', models.TextField(blank=True)),
                ('feedback_source', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='FBReporter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('org_name', models.CharField(max_length=100, unique=True, verbose_name='Organisation')),
                ('email', models.EmailField(max_length=254)),
            ],
        ),
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('source_ip', models.CharField(max_length=39)),
                ('recordcount', models.IntegerField()),
                ('policyevaluated_disposition', models.CharField(max_length=10)),
                ('policyevaluated_dkim', models.CharField(max_length=4)),
                ('policyevaluated_spf', models.CharField(max_length=4)),
                ('policyevaluated_reasontype', models.CharField(blank=True, max_length=75)),
                ('policyevaluated_reasoncomment', models.CharField(blank=True, max_length=100)),
                ('identifier_headerfrom', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('report_id', models.CharField(max_length=100)),
                ('date_begin', models.DateTimeField(db_index=True)),
                ('date_end', models.DateTimeField()),
                ('policy_domain', models.CharField(max_length=100)),
                ('policy_adkim', models.CharField(max_length=1, verbose_name='DKIM alignment mode')),
                ('policy_aspf', models.CharField(max_length=1, verbose_name='SPF alignment mode')),
                ('policy_p', models.CharField(max_length=10, verbose_name='Requested handling policy')),
                ('policy_sp', models.CharField(max_length=10, verbose_name='Requested handling policy for subdomains')),
                ('policy_pct', models.SmallIntegerField(verbose_name='Sampling rate')),
                ('report_xml', models.TextField(blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='Reporter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('org_name', models.CharField(max_length=100, unique=True, verbose_name='Organisation')),
                ('email', models.EmailField(max_length=254)),
            ],
        ),
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('record_type', models.CharField(max_length=4)),
                ('domain', models.CharField(max_length=100)),
                ('result', models.CharField(max_length=9)),
                ('record', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='dmarc.Record')),
            ],
        ),
        migrations.AddField(
            model_name='report',
            name='reporter',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='dmarc.Reporter'),
        ),
        migrations.AddField(
            model_name='record',
            name='report',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='records', to='dmarc.Report'),
        ),
        migrations.AddField(
            model_name='fbreport',
            name='reporter',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='dmarc.FBReporter'),
        ),
        migrations.AlterUniqueTogether(
            name='report',
            unique_together=set([('reporter', 'report_id', 'date_begin')]),
        ),
    ]



# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""
DMARC models for managing Aggregate Reports
https://dmarc.org/resources/specification/
"""
from django.contrib import admin

from dmarc.models import Report


class ReportAdmin(admin.ModelAdmin):
    """Report display options"""

    actions = []
    model = Report
    list_display = ['report_id', 'reporter', 'date_begin']
    list_filter = ['date_begin', 'reporter']
    readonly_fields = [
        'report_id', 'reporter',
        'date_begin', 'date_end', 'policy_domain',
        'policy_adkim', 'policy_aspf',
        'policy_p', 'policy_sp',
        'policy_pct',
        'report_xml'
    ]
    order = ['-id']

    def has_add_permission(self, request):
        return False


admin.site.register(Report, ReportAdmin)



"""Managing DMARC aggregate and feedback reports"""
from setuptools import setup
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='django-dmarc2',
    version='0.5.2',
    packages=['dmarc'],
    include_package_data=True,
    license='BSD',
    description='Managing DMARC aggregate and feedback reports',
    long_description=long_description,
    url='https://github.com/webtweakers/django-dmarc',
    download_url='https://pypi.python.org/pypi/django-dmarc2',
    author='Alan Hicks',
    author_email='ahicks@p-o.co.uk',
    requires=['django'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='dmarc email spf dkim',
)



# ----------------------------------------------------------------------
# Copyright (c) 2015-2019, Persistent Objects Ltd http://p-o.co.uk/
#
# License: BSD
# ----------------------------------------------------------------------
"""Import DMARC Feedback Reports"""
from __future__ import unicode_literals

import logging
import os
import tempfile
from argparse import FileType
from cStringIO import StringIO
from datetime import datetime
from email import message_from_string
from email.generator import Generator
from email.utils import mktime_tz, parsedate_tz

import pytz
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from dmarc.models import FBReport, FBReporter

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Command class for importing DMARC Feedback Reports
    Most errors are not raised to prevent email bounces
    """
    help = 'Imports a DMARC Feedback Report from an email'

    def add_arguments(self, parser):
        parser.add_argument(
            '-e',
            '--email',
            type=FileType('r'),
            default=False,
            help='Import from email file, or - for stdin'
        )

    def handle(self, *args, **options):
        """
        Handle method to import a DMARC Feedback Report
        """

        logger.info("Importing DMARC Feedback Report")

        if not options['email']:
            msg = "Check usage, please supply a single DMARC report file or email"
            logger.error(msg)
            raise CommandError(msg)

        msg = 'Processing email'
        logger.debug(msg)

        try:
            email = options['email'].read()
            dmarcemail = message_from_string(email)
        except:
            msg = 'Unable to use email'
            logger.debug(msg)
            raise CommandError(msg)

        if dmarcemail.is_multipart():
            self.process_multipart(dmarcemail)
        else:
            self.process_822(dmarcemail)

    @staticmethod
    def process_multipart(dmarcemail):
        """Extract multipart report"""
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        # pylint: disable=too-many-nested-blocks
        report = FBReport()
        dmarc_reporter = None
        try:
            dmarcemail.get_payload()
            dmarc_reporter = dmarcemail.get('from')
            report.reporter = FBReporter.objects.get(email=dmarc_reporter)
            mimepart = dmarcemail.get_payload()
        except ObjectDoesNotExist:
            try:
                report.reporter = FBReporter.objects.create(
                    org_name=dmarc_reporter,
                    email=dmarc_reporter,
                )
            except:
                msg = 'Failed to find or create reporter {}'.format(dmarc_reporter)
                logger.error(msg)
                raise CommandError(msg)
        except:
            msg = 'Unable to get rfc822 report'
            logger.error(msg)
            temp = tempfile.mkstemp(prefix='dmarc-', suffix='.eml')
            tmpf = os.fdopen(temp[0], 'w')
            tmpf.write(dmarcemail.get_payload())
            tmpf.close()
            msg = 'Saved as: {}'.format(temp[1])
            logger.error(msg)
            raise CommandError(msg)
        out = StringIO()
        gen = Generator(out, maxheaderlen=0)
        gen.flatten(dmarcemail)
        report.feedback_source = out.getvalue()
        gen = None
        out = None

        # Get the human readable part
        try:
            mimepart = dmarcemail.get_payload(0)
            if mimepart.get_content_type() == 'text/plain':
                # get the human-readable part of the message
                report.description = mimepart
        except:
            msg = 'Unable to get human readable part'
            logger.warning(msg)

        # Get the feedback report
        try:
            mimepart = dmarcemail.get_payload(1)
            if mimepart.get_content_type() == 'message/feedback-report':
                out = StringIO()
                gen = Generator(out, maxheaderlen=0)
                gen.flatten(mimepart)
                report.feedback_report = out.getvalue()
                gen = None
                out = None
            else:
                msg = 'Found {} instead of message/feedback-report'.format(mimepart.get_content_type())
                logger.error(msg)
        except:
            msg = 'Unable to get feedback-report part'
            logger.error(msg)

        if report.feedback_report:
            for line in report.feedback_report.splitlines():
                line = line.lstrip()
                (ls0, ls1, ls2) = line.partition(':')
                ls0 = ls0.strip()
                ls2 = ls2.strip()
                if ls1:
                    if not report.domain:
                        if ls0 == 'Reported-Domain':
                            report.domain = ls2
                    if not report.source_ip:
                        if ls0 == 'Source-IP':
                            report.source_ip = ls2
                    if not report.email_from:
                        if ls0 == 'Original-Mail-From':
                            report.email_from = ls2
                    if not report.date:
                        if ls0 == 'Arrival-Date':
                            try:
                                # get tuples
                                tuples = parsedate_tz(ls2)
                                # get timestamp
                                time = mktime_tz(tuples)
                                report.date = datetime.fromtimestamp(time)
                                tz_utc = pytz.timezone('UTC')
                                report.date = report.date.replace(tzinfo=tz_utc)
                            except:
                                msg = 'Unable to get date from: {}'.format(ls2)
                                logger.error(msg)
                                report.date = datetime.now()
                    if not report.dmarc_result:
                        if ls0 == 'Delivery-Result':
                            report.dmarc_result = ls2
                    if ls0 == 'Authentication-Results':
                        auth_results = ls2.split()
                        for result in auth_results:
                            (typ, eq_sign, alignment) = result.partition('=')
                            if not eq_sign:
                                continue
                            if not report.dkim_alignment and typ == 'dkim':
                                report.dkim_alignment = alignment.rstrip(';')
                            if not report.spf_alignment and typ == 'spf':
                                report.spf_alignment = alignment.rstrip(';')

        # Get the rfc822 headers and any message
        out = StringIO()
        gen = Generator(out, maxheaderlen=0)
        try:
            mimepart = dmarcemail.get_payload(2, False)
            mimepart_type = mimepart.get_content_type()
            gen.flatten(mimepart)
            if mimepart_type == 'message/rfc822':
                report.email_source = out.getvalue()
            elif mimepart_type == 'message/rfc822-headers':
                report.email_source = out.getvalue()
            elif mimepart_type == 'text/rfc822':
                report.email_source = out.getvalue()
            elif mimepart_type == 'text/rfc822-headers':
                report.email_source = out.getvalue()
            else:
                msg = 'Found {} instead of rfc822'.format(mimepart_type)
                logger.debug(msg)
        except:
            msg = 'Unable to get rfc822 part'
            logger.warning(msg)
        gen = None
        out = None
        if report.email_source:
            for line in report.email_source.splitlines():
                line = line.lstrip()
                (ls0, ls1, ls2) = line.partition(':')
                ls0 = ls0.strip()
                ls2 = ls2.strip()
                if ls1:
                    if not report.email_subject:
                        if ls0 == 'Subject':
                            report.email_subject = ls2

        try:
            report.save()
        except:
            msg = 'Failed save from {}'.format(report.reporter)
            logger.error(msg)
            temp = tempfile.mkstemp(prefix='dmarc-', suffix='.eml')
            tmpf = os.fdopen(temp[0], 'w')
            tmpf.write(dmarcemail.get_payload())
            tmpf.close()
            msg = 'Saved as: {}'.format(temp[1])
            logger.error(msg)

    @staticmethod
    def process_822(dmarcemail):
        """Extract report from rfc822 email, non standard"""
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        report = FBReport()
        dmarc_reporter = None
        try:
            dmarc_reporter = dmarcemail.get('from')
            report.reporter = FBReporter.objects.get(email=dmarc_reporter)
        except ObjectDoesNotExist:
            try:
                report.reporter = FBReporter.objects.create(
                    org_name=dmarc_reporter,
                    email=dmarc_reporter,
                )
            except:
                msg = 'Failed to find or create reporter {}'.format(dmarc_reporter)
                logger.error(msg)
                raise CommandError(msg)
        except:
            msg = 'Unable to get feedback report'
            logger.warning(msg)
            temp = tempfile.mkstemp(prefix='dmarc-', suffix='.eml')
            tmpf = os.fdopen(temp[0], 'w')
            tmpf.write(dmarcemail.get_payload())
            tmpf.close()
            msg = 'Saved as: {}'.format(temp[1])
            logger.error(msg)
            raise CommandError(msg)
        report.feedback_source = dmarcemail.get_payload()
        out = StringIO()
        gen = Generator(out, maxheaderlen=0)
        gen.flatten(dmarcemail)
        report.email_source = out.getvalue()
        gen = None
        out = None
        logger.info("Feedback report source: %s", report.feedback_source)

        for line in report.feedback_source.splitlines():
            line = line.lstrip()
            (ls0, ls1, ls2) = line.partition(':')
            ls0 = ls0.strip()
            ls2 = ls2.strip()
            if ls1:
                if not report.domain:
                    if ls0 == 'Sender Domain':
                        report.domain = ls2
                if not report.source_ip:
                    if ls0 == 'Sender IP Address':
                        report.source_ip = ls2
                if not report.date:
                    if ls0 == 'Received Date':
                        try:
                            # get tuples
                            tuples = parsedate_tz(ls2)
                            # get timestamp
                            time = mktime_tz(tuples)
                            report.date = datetime.fromtimestamp(time)
                            tz_utc = pytz.timezone('UTC')
                            report.date = report.date.replace(tzinfo=tz_utc)
                        except:
                            msg = 'Unable to get date from: {}'.format(ls2)
                            logger.error(msg)
                            report.date = datetime.now()
                if not report.spf_alignment:
                    if ls0 == 'SPF Alignment':
                        report.spf_alignment = ls2
                if not report.dkim_alignment:
                    if ls0 == 'DKIM Alignment':
                        report.dkim_alignment = ls2
                if not report.dmarc_result:
                    if ls0 == 'DMARC Results':
                        report.dmarc_result = ls2
                if not report.email_from:
                    if ls0 == 'From':
                        report.email_from = ls2
                if not report.email_subject:
                    if ls0 == 'Subject':
                        report.email_subject = ls2
        try:
            report.save()
        except:
            msg = 'Failed save from {}'.format(dmarc_reporter)
            logger.error(msg)
            temp = tempfile.mkstemp(prefix='dmarc-', suffix='.eml')
            tmpf = os.fdopen(temp[0], 'w')
            tmpf.write(dmarcemail.get_payload())
            tmpf.close()
            msg = 'Saved as: {}'.format(temp[1])
            logger.error(msg)
